/*
  ShipRange.cpp - Show ships according to the scanner.

  Jason Hood, 18 January, 2013.

  Normally in MP, NPC ships disappear at 2.5K, players at 5K and group members
  at 10K.  This plugin will change NPCs and players to the range of the scanner
  and group members to double the range.  Note that whilst ships will be listed
  at large distances, you still need to patch Freelancer.exe to select them
  beyond the default 10K limit.

  v1.10, 27 January, 2013.

  Allow particular mounted (even if inactive) equipment to either interfere with
  another player's scanner (i.e. a jammer) or boost your own range.  Define the
  equipment in the new DATA\EQUIPMENT\radarmoddb.ini file:

	[RadarModDB]
	radar_mod = equipment_nickname, radar_mod_value
	...

  where equipment_nickname should be obvious and radar_mod_value determines how
  the radar is affected:

	<= 0 - ignored
	<= 1 - interference: new_range = old_range * (1 - radar_mod_value)
	>  1 - boost: new_range = old_range * radar_mod_value

  If more than one type is mounted, the highest value will be used (i.e. they
  don't accumulate).  Furthermore, due to the method employed, it does not
  apply to NPCs (i.e. even if an NPC has a mounted jammer, your range will not
  be affected, nor will having a mounted jammer yourself affect an NPC's range;
  a mounted booster will not increase an NPC's range, but you will be able to
  see farther NPCs).

  Example:
       radar_mod = ge_s_cm_01, 0.33
       radar_mod = ge_s_cm_02, 1.25

  means the Countermeasure Dropper will decrease the scan range of other players
  by one third (e.g. 2500 becomes 1675), whilst the Imp. Countermeasure
  Dropper will boost your own range by one quarter (e.g. 2500 becomes 3125).

  Install:
	Copy to EXE, add to [Libraries] in dacomsrv.ini.
	Requires and assumes v1.1.
	Does *not* work in conjunction with AdminRange (whichever is installed
	 first will have precedence).
*/

#include "Common.h"
#include "Server.h"
#include <map>

#define NAKED	 __declspec(naked)
#define FASTCALL __fastcall


#define ADDR_RANGE ((PDWORD)(0x6cf7270+2))


DWORD dummy;
#define ProtectX( addr, size ) \
  VirtualProtect( addr, size, PAGE_EXECUTE_READWRITE, &dummy )
#define ProtectW( addr, size ) \
  VirtualProtect( addr, size, PAGE_READWRITE, &dummy )

#define INDIRECT( from, to ) \
  to##_Old = **(PDWORD*)(from); \
  to##_New = (DWORD)to##_Hook; \
  *(PDWORD*)(from) = &to##_New


#define pNPC_range    ((float*)0x6d66aec)
#define pPlayer_range ((float*)0x6d66af0)
#define pGroup_range  ((float*)0x6d66af4)

typedef std::map<UINT, float> RadarModDBMap;

RadarModDBMap RadarModDB;


void init()
{
  // Let's just assume "data path = ..\data".
  INI_Reader ini;
  if (!ini.open( "..\\data\\equipment\\radarmoddb.ini" ))
    return;
  while (ini.read_header())
  {
    if (ini.is_header( "RadarModDB" ))
    {
      while (ini.read_value())
      {
	if (ini.is_value( "radar_mod" ))
	{
	  UINT equip = CreateID( ini.get_value_string( 0 ) );
	  float mod = ini.get_value_float( 1 );
	  if (mod > 0)
	    RadarModDB[equip] = mod;
	}
      }
    }
  }
}


bool GetRadarMod( UINT player, float& mod, bool jammer )
{
  std::list<EquipDesc>::const_iterator iter, end;
  RadarModDBMap::const_iterator miter;
  bool found = false;
  mod = 0;

  --player;
  iter = Players.playerdata[player].equipdesclist.equip.begin();
  end  = Players.playerdata[player].equipdesclist.equip.end();
  for (; iter != end; ++iter)
  {
    if (!iter->bMounted)
      continue;
    miter = RadarModDB.find( iter->iArchID );
    if (miter == RadarModDB.end())
      continue;
    if ((jammer && miter->second <= 1) ||
	(!jammer && miter->second > 1))
    {
      found = true;
      if (miter->second > mod)
	mod = miter->second;
    }
  }
  return found;
}


void FASTCALL CheckRange( UINT player, UINT scanee )
{
  UINT	ship = 0;
  int	radar = 2500, cargo;
  float range, mod;

  pub::Player::GetShip( player, ship );
  pub::SpaceObj::GetScannerRange( ship, radar, cargo );

  range = radar;
  if (GetRadarMod( player, mod, false ))
    range *= mod;
  if (scanee && GetRadarMod( scanee, mod, true ))
    range *= 1 - mod;

  *pNPC_range	 =
  *pPlayer_range = range * range;
  *pGroup_range  = (range * 2) * (range * 2);
}


DWORD Range_Old, Range_New;

NAKED
void Range_Hook()
{
  __asm {
	mov	edx, [esi+0x10]
	mov	edx, [edx+0xb4]
	mov	ecx, [edi+0x38]
	call	CheckRange
	jmp	Range_Old

	align	16
  }
}


void Patch()
{
  init();

  ProtectX( ADDR_RANGE,    4 );
  ProtectW( pNPC_range,    4 );
  //ProtectW( pPlayer_range, 4 );
  //ProtectW( pGroup_range,  4 );

  INDIRECT( ADDR_RANGE, Range );
}


BOOL WINAPI DllMain( HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved )
{
  if (fdwReason == DLL_PROCESS_ATTACH)
    Patch();

  return TRUE;
}
